/***************************************************************************
Copyright (c) 2006 Peter Korzuszek, Damian Rakowski, KateOS Project
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 
*Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
 
*Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE. 
****************************************************************************/

#ifndef SMART_TOOLS_VALUE_LIST_H
#define SMART_TOOLS_VALUE_LIST_H

#include "stsharedmemory.h"
#include "smarttools.h"
#include "stassert.h"
#include "stdebug.h"
#include <stdlib.h>
#include <string>
#include "stoutput.h"

/*!
	\brief StValueList allows you to keep variable table of any type.
	
	It works very similiarly to the vector class from STD library but doesn't
	have iterators yet. If you want to create StValueList<StString>, the
	better way is to use the StStringList class, which allows you to load
	and save lines of your strings into files.
*/

template <class T>
class StValueList
{
	struct Element{
		T data;
		Element* pointer;
		Element( const T& t ) : data(t) {  }
		
		void* operator new( size_t, void* loc ){
			return loc;
		}
	};
	
	public:
		StValueList();
		//StValueList( const StValueList<T>& copy ){ init(); this->copy( copy ); }
		/**
			Create and fill StValueList with default values.
			\sa fill()
		*/
		StValueList( unsigned count, const T& item ) { init(); fill(count, item); }
		~StValueList();
		
		/** Insert copy of item at the end of list */
		void push_back( const T& item );
		/** Delete first item */
		void pop_back();
		
		/** Insert copy of item at the begining of list */
		void push_front( const T& item );
		/** Delete last item */
		void pop_front();
		
		/** Adds item only if new in list */
		void addUnique( const T& item );
		/** Adds only these list items that are new to this list */
		void addUnique( const StValueList<T>& list );

		/**
			Fills list with 'count' values equals item.
			It's usefull when you want blank value list with
			default value. It does not delete od StValueList contents.
		*/
		void fill( unsigned count, const T& item );
		
		/** \return index of first occurance of item */
		int find( const T& item ) const;
		/** \return number of occurances of item */
		int contains( const T& item ) const;
		
		/*!
			Delete all items that are equal to item
			\return num of deleted items
		*/
		int remove( const T& item );
		/** Remove item on index */
		void remove( unsigned index );
		
		/** Clear whole list */
		void clear();
		
		/** \return item that is hold on index */
		const T& at( unsigned index ) const;
		/**
			Version of function above with modify permissions.
			\return item that is hold on index
		*/
		T& at( unsigned index );

		/** \return the first element of list */
		const T& first() const { ST_ASSERT( size() > 0 ); return at(0); }
		/**
			Version of first() with modify permissions.
			\return the first element of list
		*/
		T& first() { ST_ASSERT( size() > 0 ); copy(); return at(0); }

		/** \return the last element of list */
		const T& last() const { ST_ASSERT( size() > 0 ); return at(size()-1); }
		/**
			Version of last() with modify permissions.
			\return the last element of list
		*/
		T& last() { ST_ASSERT( size() > 0 ); copy(); return at(size()-1); }
		
		/** \return number of list items */
		unsigned size() const { return itemCount; }
		/** \return true if there is no items in list */
		bool empty() const { return ( itemCount==0 ) ? true : false; }
		
		StValueList& operator<<(const T& item){ push_back(item); return *this; }
		StValueList& operator<<( const StValueList<T>& list );
		
		T& operator[](unsigned index) { copy(); return at(index); }
		const T& operator[](unsigned index) const { return at(index); }
		bool operator==( const StValueList& list ) const;
		
		//bool operator=( const StValueList<T>& copy );
	
	protected:
		StSharedMemory memory;
		unsigned itemCount;
		
		void elementCreate( const T& data, void* placeToCopy );
		void elementDestroy( void* memoryPointer );
		
		void copy( void* mem, unsigned itemCount );
		void copy( const StValueList<T>& copy );
		void copy();
	
	private:		
		friend class StStringList;
		
		void init();
};

// -------------------Constructors-------------------------
template <class T>
inline StValueList<T>::StValueList()
{
	init();
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::init()
{
	itemCount = 0;
}

// -------------------Destructor---------------------------

template <class T>
inline StValueList<T>::~StValueList()
{
	if( memory.getInstances() == 1 )
		clear();
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::elementCreate( const T& data, void* placeToCopy )
{	
	new(placeToCopy) Element( data );
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::elementDestroy( void* memoryPointer )
{	
	((Element*)memoryPointer)->Element::~Element();
}

// --------------------------------------------------------

template <class T>
void StValueList<T>::push_back( const T& item )
{
	// this will change memory content
	copy();
	
	unsigned memorySize = memory.getMemorySize() + sizeof(Element);
	memory.realloc( memorySize );
	
	elementCreate( item, (void*)&((char*)memory.getPointer())[sizeof(Element)*itemCount] );
	
	itemCount++;
}

// --------------------------------------------------------

template <class T>
void StValueList<T>::push_front( const T& item )
{
	// this will change memory content
	copy();
	
	unsigned memorySize = memory.getMemorySize() + sizeof(Element);
	memory.realloc( memorySize );
	
	// first of all I should move all data
	memmove( (void*)&((char*)memory.getPointer())[sizeof(Element)], memory.getPointer(), sizeof(Element)*itemCount );
	
	elementCreate( item, memory.getPointer() );
	itemCount++;
}

// --------------------------------------------------------

template <class T>
int StValueList<T>::find( const T& item ) const
{
	Element *e;
	
	for( unsigned i=0; i < itemCount; i++ ){

		e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*i];
		
		if( e->data == item )
			return i;
	}
	
	return -1;
}

// --------------------------------------------------------

template <class T>
void StValueList<T>::remove( unsigned index )
{
	ST_ASSERT( index < itemCount );
	
	// this will change memory content
	copy();
	
	elementDestroy( (void*)&((char*)memory.getPointer())[sizeof(Element)*index] );
	
	// if there is more items further in memory I should copy these one position back
	if( index+1 < itemCount )
		memmove( (void*)&((char*)memory.getPointer())[sizeof(Element)*index], (void*)&((char*)memory.getPointer())[sizeof(Element)*(index+1)], sizeof(Element)*(itemCount-(index+1)) );
	
	// resizing memory
	unsigned memorySize = memory.getMemorySize() - sizeof(Element);
	memory.realloc( memorySize );
	
	itemCount--;
}

// --------------------------------------------------------

template <class T>
int StValueList<T>::contains( const T& item ) const
{
	int count = 0;
	Element *e;
	
	for( unsigned i=0; i < itemCount; i++ ){
		e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*i];
		
		if( e->data == item )
			count++;
	}
	
	return count;
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::fill( unsigned count, const T& item )
{
	for( unsigned i=0; i < count; i++ )
		push_back( item );
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::pop_back()
{
	ST_ASSERT( itemCount > 0 );
	
	remove( size()-1 );
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::pop_front()
{
	ST_ASSERT( itemCount > 0 );
	
	remove(0);
}



// --------------------------------------------------------

template <class T>
int StValueList<T>::remove( const T& item )
{
	int count=0;
	int index = -1;
	
	do{
		if( index != -1 ){
			remove(index);
			
			count++;
			//itemCount--;
		}
		
		index = find( item );
	}while( index != -1 );
	
	return count;
}

// --------------------------------------------------------

template <class T>
inline const T& StValueList<T>::at( unsigned index ) const
{
	ST_ASSERT( index < itemCount );	
	Element* e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*index];	
	return e->data;
}

// --------------------------------------------------------

template <class T>
inline T& StValueList<T>::at( unsigned index )
{
	ST_ASSERT( index < itemCount );
	
	copy();
	
	Element* e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*index];	
	return e->data;
}

// --------------------------------------------------------

template <class T>
void StValueList<T>::clear()
{
	// this will change memory content
	copy();
	
	for( unsigned i=0; i < size(); i++ ){
	
		elementDestroy( (void*)&((char*)memory.getPointer())[sizeof(Element)*i] );
	}
	
	memory.free();
	init();
}

// --------------------------------------------------------

template <class T>
inline StValueList<T>& StValueList<T>::operator<<( const StValueList<T>& list )
{
	for( unsigned i=0; i < list.size(); ++i )
		push_back( list[i] );
	
	return *this;
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::addUnique( const T& item )
{
	if( find( item ) == -1 )
		push_back(item);
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::addUnique( const StValueList<T>& list )
{
	for( unsigned i=0; i < list.size(); i++ )
		addUnique( list[i] );
}

// --------------------------------------------------------

template <class T>
void StValueList<T>::copy( const StValueList<T>& copy )
{
	// copy from another object
	
	// then copy memory with objects
	this->copy( copy.memory.getPointer(), copy.itemCount );
}

// --------------------------------------------------------

template <class T>
void StValueList<T>::copy( void* mem, unsigned itemCount )
{
	// be extremaly carefull with that!
	
	memory.renew();
	memory.malloc( itemCount * sizeof(Element) );
	
	for( unsigned i=0; i < itemCount; ++i ){
		elementCreate( *(T*)&((char*)mem)[i*sizeof(Element)], (void*)&memory[i*sizeof(Element)] );
	}
	
	this->itemCount = itemCount;
}

// --------------------------------------------------------

template <class T>
inline void StValueList<T>::copy()
{
	// make a copy of the same object
	// only if instance is grather than 1
	
	if( memory.getInstances() > 1 ){	
		// then let's make copy
		copy( memory.getPointer(), itemCount );
	}
}

// --------------------------------------------------------

template <class T>
bool StValueList<T>::operator==( const StValueList& list ) const
{
	if( this->size() != list.size() ) return false;
	
	int size = this->size();
	for( int i = size; i < size; i++ )
			if( at( i ) != list.at( i ) ) return false;
	
	return true;
}

// --------------------------------------------------------


#endif //SMART_TOOLS_VALUE_LIST_H
